/* 17.08.2002 C Bauer (Berlin)  
 *
 * root solving routine taken from "Numerical Recipies in C" /
 * ( W. H. Pres et. al., Cambr. Univ. Press, 1992)
 *
 *
 * hermite mapping table generator
 *
 * add the output to the function arguments of the respective
 * cal-files which are prepared for hermite mapping
 *
 *
 * nb - !! this program was written in the well known "chaotic mode" 
 *         i.e. starting out with an idea, seeing that it does not work
 *         as planned, then adding successively all that is needed to make
 *         it work...
 *         so don't look at the code too critically...
 */
 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>


double p0, p1, r0, r1;
double p10, p11, r10, r11;
int res, resint;
double a, b, c, d, dmod;
double a1, b1, c1;
double A, B, C, D, E;
double delta, deltaint, tdeltaint, td2, tl, th;
double xev;
double rsmult, rsacc;
double integ[3];
double xpoints[3];
double ypoints[3];
double ipcoefs[3];
double** ipcflist;

int i, j;
char oname[256];
FILE *outstream;

double *len;
double *lenint;
double *tvals;

void printhelp();
double xhermite(double);
double poly(double);
double rtflsp( double (*func)(), double, double, double);
void genipcfs(double*, double*, double*);

int
main(int argc, char** argv)
{

if(argc < 10){
    if(argc == 1)
        fprintf(stderr, "too few arguments, type genmaptab -h for help\n");
    else{
     if (!strcmp(argv[1], "-h") || !strcmp(argv[1], "-help"))
                        printhelp();           
     else
     fprintf(stderr, "wrong number of arguments, type genmaptab -h for help\n");
    }
    exit(0);
        
}
  /* the mandatory arguments */
p0 = atof(argv[1]); p1 = atof(argv[2]); r0 = atof(argv[3]); r1 = atof(argv[4]);
p10 = atof(argv[5]); p11 = atof(argv[6]); r10 = atof(argv[7]); r11 = atof(argv[8]);
res = atoi(argv[9]);

 /* the defaults */
rsmult = 3.5;
rsacc = 0.01; 

for(i=10; i<argc; i++){
  if (argv[i] == NULL)
             break;
  if(argv[i][0] != '-'){
      fprintf(stderr, "error in option settings");
      exit(0);
  }
  switch(argv[i][1]){
    case 'a': rsacc = atof(argv[++i]);
              break;
    case 'f': rsmult = atof(argv[++i]);
              break;
  }
}  	      	            

resint = res*2; /* doubled internal resolution for integration */

len = (double*)malloc((res+1)*sizeof(double));
lenint = (double*)malloc((resint+1)*sizeof(double));
tvals = (double*)malloc((resint+1)*sizeof(double));
ipcflist = (double**)calloc(res,sizeof(double*));
for(i=0; i<res; i++)
        ipcflist[i] = (double*)malloc(3*sizeof(double));

a = 2*p0 - 2*p1 + r0 +r1;
b = 3*p1 - 3*p0 -2*r0 - r1;
c = r0;
d = p0;
dmod = p0;
a1 = 2*p10 - 2*p11 + r10 +r11;
b1 = 3*p11 - 3*p10 -2*r10 - r11;
c1 = r10;


                    /*prepare to get the tvals */
deltaint = (p1-p0)/(float)resint;

tdeltaint = 1.0/(float)resint;
td2 = rsmult*tdeltaint;
       
                  /* determine parameter vals for the x divisions */
for(i=1; i<resint; i++){
       xev = deltaint*i;
       dmod = d - xev;   /*modify function */
       
       tl = tdeltaint*i - td2;
       th = tdeltaint*i + td2;
       if(tl < 0.0) tl = 0.0;
       if(th > 1.0) th = 1.0; 
       
       tvals[i] = rtflsp(xhermite, tl, th, rsacc); 
}
tvals[0] = 0.0;
tvals[resint] = 1.0;

              /* the curve -length polynom coeffs */
A = 9*(a*a+a1*a1); 
B = 12*(a*b+a1*b1);
C = 6*(a*c+a1*c1)+4*(b*b+b1*b1);
D = 4*(b*c+b1*c1);
E = c*c + c1*c1;

lenint[0] = 0.0;
           /* stepwise curve length integral*/

for(i=1; i<= resint; i++){

     integ[0] = poly(tvals[i-1]);
     integ[1] = poly(0.5*(tvals[i]+tvals[i-1]));
     integ[2] = poly(tvals[i]);
          
     lenint[i] = integ[0]/3 + 4*integ[1]/3 + integ[2]/3;
     lenint[i] = lenint[i-1]+0.5*(tvals[i]-tvals[i-1])*lenint[i];
     
}
           /* use only every second val - switch back to external res */
j=0;  
for(i = 0; i<=resint; i+=2){
   len[j] = lenint[i];   
   j++;
}
delta=(p1-p0)/(float)res;   /* the external delta */    

            /* determine coeffs for quadratic interpolation */
                                      
for(i=1; i<res; i++){

 xpoints[0] = (i-1)*delta; xpoints[1] = i*delta; xpoints[2] = (i+1)*delta;
 ypoints[0] = len[i-1]; ypoints[1] = len[i]; ypoints[2] = len[i+1];

 genipcfs(ipcflist[i], xpoints, ypoints);
} 

printf("%f %f %f\n", p0, delta, p10);

for(i = 1; i<res; i++)
  printf("%f %f %f \n", ipcflist[i][0], ipcflist[i][1], ipcflist[i][2]);
 
printf("## %i additional params\n", 3*res);
printf("## total curve length: %f\n", len[res]);
 
}
     
double
xhermite(double t)
{
return(a*t*t*t + b*t*t + c*t +dmod);
}

double
poly(double tev)
{
 double val;
 val = A*tev*tev*tev*tev + B*tev*tev*tev + C*tev*tev + D*tev + E; 
 return(sqrt(val));
} 

double
rtflsp(double (*func)(double), double x1, double x2, double xacc)
{
int j;
double fl, fh, xl, xh, swap, dx, del, f, rtf;

fl = (*func)(x1);
fh = (*func)(x2);

if (fl*fh > 0.0) {
   fprintf(stderr, "root solving error, try higher -f option, (currently: f = %3.3f)\n", rsmult);
   exit(0);
}
if(fl < 0.0){
  xl = x1;
  xh = x2;
} else {
  xl = x2;
  xh = x1;
  swap = fl;
  fl = fh;
  fh = swap;
}
dx = xh -xl;
for (j = 1; j<=30; j++) {
  rtf = xl+dx*fl/(fl-fh);
  f = (*func)(rtf);
  if (f <0.0) {
    del = xl -rtf;
    xl = rtf;
    fl = f;
  } else {
    del = xh-rtf;
    xh = rtf;
    fh = f;
  }
  dx = xh -xl;
  if (fabs(del) < xacc || f == 0.0 ) return(rtf);
}  
  fprintf(stderr, "root solving error");
  return(0.0);
}                 
     
void
genipcfs(double* cfs, double* xps, double* yps)
{

double K0, K1, K2;

K0 = yps[0]/((xps[0]-xps[1])*(xps[0]-xps[2]));
K1 = yps[1]/((xps[1]-xps[0])*(xps[1]-xps[2]));
K2 = yps[2]/((xps[2]-xps[0])*(xps[2]-xps[1]));

cfs[0] = K0 + K1 + K2;
cfs[1] = K0*(xps[1]+xps[2]);
cfs[1] += K1*(xps[0]+xps[2]);
cfs[1] += K2*(xps[0]+xps[1]);
cfs[1] = -cfs[1];
cfs[2] = K0*xps[1]*xps[2];
cfs[2] += K1*xps[0]*xps[2];
cfs[2] += K2*xps[0]*xps[1];

return;
}     
     
     
     
void
printhelp()
{
printf("\n");
printf("Usage:\n");
printf("genmaptab Hpx0 Hpx1 Hrx0 Hrx1 Hpy0 Hpy1 Hry0 Hry1 res [ -f val -a val] \n");
printf("with:\n");
printf("Hpx0 ... Hrx1  Hpy0 ... Hry1 - hermite parameters from the object definition\n");
printf("res - number of divisions (cal-files are prepared for a maximum of 20 steps)\n");
printf("\n");
printf("The y above has only dummy status, it may be z as well.\n");
printf("The first hermite defines the direction, along which the curve\n");
printf("proceeds, and this has to be the x-direction.\n\n"); 
printf("Parameters Hpx0 until res are mandatory, the following are optional:\n");
printf("option -f - factor needed in the root-solving process, default 3.5, set it\n");
printf("            to a higher value if needed\n");
printf("option -a - accuracy threshhold for root solving, default: 0.01\n");
printf("\n"); 
}
         