/* Copyright (c) 29.07.2002 Carsten Bauer (Berlin)  */

/*    lite.c - experimental module for Radiance  */

/*
 *    3-d storing and reusing direct illuminance values
 *
 *
 */    
 
 
#include "ray.h"
#include "octree.h"
#include "otypes.h"


#include "lite.h"


extern int ambounce;
short liflags = 1;
short dctrig;


static int bfcnt = 0;
static int tvtrig;
static int *newsind;
static float *brlv;
static int read = 0;
static char maxdep;
static char sdep;

static float cvarr[16] = {1.0,1.0,1.0,1.0,1.0,1.0,1.0,0.5,0.32,0.2,0.125,
                   0.125,0.125,0.125,0.125,0.125};
static short rdarr[6];

static float delpos[4];

/*  functions */

void initlite(char*);
void closelite();

void checkfile(char*);
void readfile(FILE*);

void saveliteval(int,short,char,LITEVAL*);
int insertlite(int,short,char,char*);

int checklite(COLOR,RAY*,FVECT,int*,short*,double);
int setlite(COLOR,FVECT,int,short,int);
int setdrclite(COLOR,RAY*,FVECT);

int initlvstore(int,char);
int addlvstore(int,char);

int getbindex(RAY*);
short getsindex(float*, char);

int reablk(int, char, int);

void
initlite(char* lfile)                   /* initialize liteval-calculation*/
{ 
  register int i, k;   
  float cv;
  float dum;
  char msg[256];
   
                              /* set the controls */			      
  switch(dcaccur){
  case 'p':
     dctrig = 0xff-2;  break;
  
  case 'n':
     dctrig = AMBIENT;  
     if(ambounce == 0){
               liflags =0;
	       return;
     }
     break;
  
  case 'o':
     liflags = 0; return; break;
   
  default:              
     error(USER, "DC: invalid -Da setting.");
  }
  

  dum = thescene.cusize;
  
  if(dccub < FTINY){
  
    if(dum < 30.0)         /* heuristic startvals for the dc cube sizes */ 
              dccub = 0.1;
    else if(dum < 100.0)
                  dccub = 0.15;
         else     dccub = 0.002 * thescene.cusize;	    	    	    
  } 

  k=0;                   /* get start refining depth */
  while(dum > dccub){
     k++;
     dum*=0.5;
  }
  if(k>10){
      sprintf(msg, "DC: specified resolution cannot be reached,\n DC: using %3.4f instead\n", 
                      thescene.cusize/(1<<10));
      fprintf(stderr, msg);		      
      k = 10;
  }
  if(k<3)
     k=3;
  dcmaxdep = k;
  if(dcsubd == 0)            /* set defaults */
       dcsubd = 2;
  else if(dcsubd < 1 || dcsubd > (dctval>0?2:5))
      error(USER, "DC: invalid blockdepth");       
  
   dcbld = dcmaxdep-dcsubd;           /* the blockdepth */
   maxdep = dcsubd+3;
 
  if(dconv < FTINY){
      for(i=0; i<6; i++)
           rdarr[i] =(int)(cvarr[dcbld+i]*aloc[i]);
  }	   
  else{  
      if(dconv > 1.0)
           error(USER, "DC: invalid red. factor");              	
      for(i=dcsubd; i<6; i++){
           rdarr[i] =(int)(dconv*aloc[i]);
	   if(rdarr[i] <= 0)
	        error(USER,"DC: invalid -Dc setting (too low)");
      }		
  } 
 
  if(lfile != NULL){
            liflags |= FILEWRT;
	    checkfile(lfile);
  }
  	    				  

  
  if(dcssampexc > FTINY){ 
      if(dcssampexc > 1.0)
         error(USER, "DC: invalid -Ds setting.");
      else	 
         liflags |= SSAMPEXC;           
  }  
  else if(dcssampexc < -FTINY)
                     error(USER, "DC: invalid -Ds setting.");
     
  liblkptr = (LITEBLOCK**)malloc(aloc[dcbld]*sizeof(LITEBLOCK*));
      for(i=0; i<aloc[dcbld]; i++)
                *(liblkptr+i) = NULL;

#ifdef INFO
  totmem+= aloc[dcbld]*4;
  maxref = dcmaxdep;
#endif
		
  if(read)
        readfile(lvstream);
                                  /* allocate writebuffer */	
  lvbufbsz = LVSBUFSZ*(LITEVALSIZ+8);           
  lvsfbuf = (char*)malloc(lvbufbsz);                       

  if(dcrlen < FTINY)             /* min raylength for nearfield exclusion */
     minrlen = 1.8*thescene.cusize/(1<<dcmaxdep);
  else
     minrlen=dcrlen;
     
                                 
  if(dctval < 0 || dctval == 1 )
      error(USER, "DC: invalid number of testvals.");     
     
  if(dctval == 0){
          sdep = dcsubd | 020;    /* a workaround */
	  dctval = 2;
  }
  else{
     if(dctval > rdarr[dcsubd]/2)
              dctval = rdarr[dcsubd]/2;
     sdep = dcsubd;	          
     tvtrig = dctval-1;
  }
  newsind = (int*)malloc(dctval*sizeof(int));
 
  dcnormdev= cos(dcnormdev*PI/180.0);       
}  

void
checkfile(char* lvname){
 
  short v2;
  float v1;
  char emg[256]; 
   
  if((lvstream=fopen(lvname,"r+")) != NULL){
     fread(&v1,1,4,lvstream);
     fread(&v2,1,2,lvstream);
          
     if(fabs(v1-dccub) > 0.01 || v2 != dcsubd ){
         
        sprintf(emg, "DC: resolution parameter inconsistency, file settings are:\n");
	fprintf(stderr,emg);
	sprintf(emg, "    -Dr %3.4f -Db %i \n", v1, v2);
	fprintf(stderr,emg);
        error(USER, "");
     } 
     read = 1;
  } 
  else{               /* open file and write res. params */
       if ((lvstream=fopen(lvname,"w+")) == NULL)
                   error(SYSTEM,"cannot open litevalfile");
       fwrite(&dccub,1,4,lvstream);
       fwrite(&dcsubd,1,2,lvstream);
       read = 0;	  
	  
  }	  	
}
 

void 
readfile(FILE* lvstr)
{        
  char cl[LITEVALSIZ];
  int bind;
  short sind;
  char bd;
  int i;
  short dum;
  int bytesread = 0;              
  int pos;
  int numlvals = 0; 
    	 
     pos = ftell(lvstr);
     while(fread(&bind,1,4,lvstr)==4){
       fread(&sind,1,2,lvstr);
       fread(&bd,1,1,lvstr);
       dum=fread(cl,1,LITEVALSIZ,lvstr);
       if(dum==LITEVALSIZ){
           bytesread+=dum;
           if(insertlite(bind,sind,bd,cl))
	            numlvals++;
       }
       else{
       pos+=bytesread;
       sprintf(errmsg,"corrupted litevalfile, ignoring all after n= %ld",
                 bytesread/LITEVALSIZ);
       error(WARNING,errmsg);
       fseek(lvstr,pos,0);
       ftruncate(fileno(lvstr),pos); 
       break;                    
       }
     }                 
     if(bytesread/LITEVALSIZ != numlvals){
            sprintf(errmsg, "DC: insert of %i litevals failed",
                  bytesread/LITEVALSIZ - numlvals); 
            error(WARNING,errmsg); 
     }      
}

void
closelite()
{
 register int j, i;                /* store a direct illuminance */
 short num;
 LITEVAL *lvp;
 LITEBLOCK *lbk;

if(liflags & FILEWRT){  /* close open blocks */
 
  for(i=0; i<aloc[dcbld]; i++){
    if((lbk=liblkptr[i]) != NULL){
          if((lbk->bldep & 020) == 0){
	          lvp = lbk->lvarr+dctval;
	          lbk->bldep |= 020;                   
                  for(j=0; j<=lbk->nlvs; j++)     /* use cached indices */
                       saveliteval(i,(short)((lvp+j)->val[0]),lbk->bldep,lbk->lvarr+j);    
          }
    }	  
  }

                    /*flush the buffer */
 fwrite(lvsfbuf,1,bfcnt,lvstream);
 fclose(lvstream);
}

#ifdef INFO	 
if(liflags){
sprintf(errmsg, "DC - values set: %i , reused %i times\n", setlvs, usedlvs );
fprintf(stderr, errmsg);
sprintf(errmsg, "DC - %i blocks (init. size %i), %i blocks realloc., tot. memory: %i bytes\n",
                                blockalc, rdarr[dcsubd], blocksrealc, totmem);
fprintf(stderr, errmsg);
sprintf(errmsg, "DC - subcube sizes %3.4f - %3.4f ( ref. depths %i - %i )\n",  
    thescene.cusize/(1<<dcmaxdep), thescene.cusize/(1<<maxref), dcbld+dcsubd, maxref );
fprintf(stderr, errmsg);

}
#endif

}

void
saveliteval(int bi, short si, char bd, LITEVAL *lvp)
{
   register int i;
   char *tmpt; 
                               /* bytewise writing into savebuffer*/    
  tmpt = (char*)&bi;
  for(i=0; i<4; i++) 
    *(lvsfbuf+bfcnt++) = *(tmpt++);
           
  tmpt = (char*)&si;
  for(i=0; i<2; i++) 
    *(lvsfbuf+bfcnt++) = *(tmpt++);

  *(lvsfbuf+bfcnt++) = bd;

  tmpt = (char*)lvp;
  for(i=0; i<LITEVALSIZ; i++)
    *(lvsfbuf+bfcnt++) = *(tmpt++);
                              
                              /* write buffer to file */  
  if(bfcnt >= lvbufbsz){      
      fwrite(lvsfbuf,1,bfcnt,lvstream);
      bfcnt = 0;
  }      
}  


int
insertlite(int bind, short sind, char bd, char *clvp)          /* insert liteval into its place */ 
{ 
  register int i, bli, sbi;
  LITEVAL  *lv;
  LITEBLOCK *lbk;
  short count;
  
 if( (dcsubd <= (bd & 017) <= 5)        && 
     (0 <= bind < aloc[dcbld])  &&
     (0 <= sind < aloc[(bd & 017)]  )       ){  

 if((lbk = liblkptr[bind]) == NULL){
      if(initlvstore(bind, bd ) == 0)
                 return(0);
      lbk = liblkptr[bind];
      count = 0;       /* first val */	       
 }
 else{
  if(lbk->indblock[sind] != -1)    /* val already present */              
        return(0);            
  count = ++(lbk->nlvs);		        
  if(count  >= lbk->maxlvs)
            if(addlvstore(bind, bd) == 0)                                                                                                                                        
	                    return(0);
 }                                                                 
#ifdef INFO
if(dcbld+(bd & 017) > maxref)
     maxref = dcbld + (bd & 017);
#endif       		

 lv = lbk->lvarr+count;    
                                /* bytewise filling into list */
 for(i=0; i<LITEVALSIZ; i++)
     *((char*)lv+i) = *(clvp++);
                               /* set octree index */
 lbk->indblock[sind] = count; 

 return(1);
 }
 else
  return(0);

}                                  

int                          
getbindex(RAY *r)         /* find octree block index for intersec. point */   
{                                    /* ip has to be within bounding cube */
   FLOAT lco[3];        /* the local coordinate */
   FLOAT lcorg[3];
   
   register int br, i, j;
   FLOAT sz;
   int bind;
   
   bind = 0;
  
   lco[0] = r->rop[0] - thescene.cuorg[0];  /* coo resp. to scenecube */
   lco[1] = r->rop[1] - thescene.cuorg[1];
   lco[2] = r->rop[2] - thescene.cuorg[2];
   
   lcorg[0] = lcorg[1] = lcorg[2] = 0.0;
   sz = thescene.cusize;
      
   i = 0;
   while(i < dcbld){             /* refine - blockind*/
      sz*=0.5;
      br = 0; 
      for(j=0; j<8; j++){
        if(lco[0] >= lcorg[0]+sz){  lcorg[0]+=sz;  br |=1;        
        }
        if(lco[1] >= lcorg[1]+sz){  lcorg[1]+=sz;  br |=2;
        }
        if(lco[2] >= lcorg[2]+sz){  lcorg[2]+=sz;  br |=4;
        }
      }
      bind = (bind<<3)+br;
      i++;
   } /* end while */
   for(j=0; j<3; j++)              /* remaining position vector in block */
       delpos[j] = (float)(lco[j]-lcorg[j]); 
   delpos[3] = sz;
   return(bind);
}

short 
getsindex(float* pos, char dp)
{
   float bcorg[3];
   register int br, i,j;
   float sz;
   short sind;
   
   sind = 0;
   bcorg[0] = bcorg[1] = bcorg[2] = 0.0;  
   sz=delpos[3];         /* remains the same, so take it from global var */
          
   
   i=0;
   while(i < dp){             /* refine -subind */
      sz*=0.5;
      br = 0; 
      for(j=0; j<8; j++){
        if(pos[0] >= bcorg[0]+sz){ bcorg[0]+=sz; br |=1;        
        }
        if(pos[1] >= bcorg[1]+sz){ bcorg[1]+=sz; br |=2;
        }
        if(pos[2] >= bcorg[2]+sz){ bcorg[2]+=sz; br |=4;
        }
      }
      sind = (sind<<3)+br;
      i++;
    } /* end while */

 return(sind); 
}      

   
int
checklite(COLOR lcol, RAY *r, FVECT nrm,           /* the lookup function of the direct cache*/ 
          int *binpt, short* sinpt, double rlen)     /* to be called with zero color !! */
{                             
 register int j;
 short lvnumb;
 register LITEVAL *lvp;
 register LITEBLOCK *lbk;
 register double ddum;          

char* errstr;
                                
 if((r->alev & 017) > 0 ){     /* handle exeptions */ 

   if(liflags & SSAMPEXC)           /*supersampl. exclusion*/ 
           if(r->alev & 020)          
                      return(2);

#ifdef IARGS 
   switch(r->parent->ro->oflags & 07){
    case 2:
      return(1);             /* complete exclusion */
    break;  

    case 1:
      if(rlen < minrlen)      /* nearfield exclusion */
                return(2);
    break;		
    default:
    break;	
   }
#endif
   
 }
#ifdef IARGS      	
 if(r->ro->oflags & 010)       /* object exclusion */                                                       
          return(2);         
#endif	  
 
 *binpt = getbindex(r);  /* get blockindex */
 
 if((lbk = liblkptr[*binpt]) != NULL){          /* block exists ? */

   *sinpt = getsindex(delpos, (lbk->bldep & 017));   /* get subblockindex */
   
   if((lvnumb= lbk->indblock[*sinpt]) != -1){      /* value in subblock exists ? */ 
    
       lvp = lbk->lvarr+lvnumb;
       ddum = 0;
       for(j=0; j<3; j++)           /* consider normal direction*/
               ddum+=lvp->dir[j]*nrm[j];
	      
       if(ddum > dcnormdev){           /* skip those with opposite normdir and perp. nd*/
           copycolor(lcol,lvp->val);
#ifdef INFO	   
           usedlvs++; 
#endif	   
           return(3);             /* found and usable*/                 
       } 
        else
	 return(2);              /* found but not usable */
   }
   return(1); /* no value found */ 
 }   
 return(0);   /* first val, new block needed */
}              
   
int
setlite(COLOR lcol, FVECT nrm, int bind, short sind, int disc)
{            
 register int j, i, k;                /* store a direct illuminance */
 short count;
 LITEVAL *lvp;
 LITEVAL *lvp0;
 LITEVAL *lvp1;
 LITEBLOCK *lbk;
 
 float rdiff;
 char dep;
  
 switch(disc){
 
 case 1:                 /* new val, block exists */
    lbk=liblkptr[bind];
    count = ++(lbk->nlvs);       
                    
    if( (lbk->bldep & 020) == 0 ){        /*testval*/
                        
       lvp=lbk->lvarr + count;   /*insert*/                                
       for(j=0; j<3; j++)
           lvp->dir[j] = (float)nrm[j];
       copycolor(lvp->val,lcol);
       lbk->indblock[sind] = count;    
#ifdef INFO
       setlvs++;
#endif     
       lvp+= dctval;                 /* cache position/index/brightness in val-array */
       for(j=0; j<3; j++)           
          lvp->dir[j] = delpos[j];
       (short)(lvp->val[0]) = sind;
       lvp->val[1] = 0.265*lcol[0] + 0.670*lcol[1] + 0.065*lcol[2];
       
       lvp0 = lbk->lvarr+dctval;
       lvp1 = lvp0+1;
       lvp0->val[2] += fabs(lvp->val[1] - lvp0->val[1]);
       lvp1->val[2] += lvp->val[1]; 
      
       rdiff= lvp0->val[2] / lvp1->val[2];
       if(rdiff > dcthresh[0]){           /* det. adequate depth */
           if(rdiff < dcthresh[1])       
                         dep=dcsubd+1;		
           else if(rdiff < dcthresh[2]) 
                              dep=dcsubd+2;
                else          dep=dcsubd+3;
       }
       else
         dep = dcsubd;	  

#ifdef INFO
if(dcbld+dep > maxref)
     maxref = dcbld + dep;
#endif       		

       if(dep > (lbk->bldep & 017)){        /* new block needed */
          
          if(reablk(bind, dep, count)){
	   
	   lvp0 = lbk->lvarr+dctval;        /* refresh pointer */
    	   for(j=0; j<=count; j++){             /* set new indices, use cached positions */ 
               newsind[j] = getsindex((lvp0+j)->dir, dep);
               lbk->indblock[newsind[j]] = j; 
           }
           if(count >= tvtrig || dep==maxdep){  /* close block */
               lbk->bldep |= 020;
    	       if(liflags & FILEWRT){                /* save testvals */                   
                  for(j=0; j<=count; j++)
                         saveliteval(bind,newsind[j],lbk->bldep,lbk->lvarr+j);
	       }
	   }
	   else{                         /* cache new indices */
    	        for(j=0; j<=count; j++)
	            (short)((lvp0+j)->val[0]) = newsind[j];
           }
	  }
	  else{          /* insufficient memory, block remains unchanged */
	   lbk->bldep |= 020; /* close it */
	  }  		
	}
	else{                         /* blockdepth remains unchanged */
	   if(count >= tvtrig){
	       lbk->bldep |= 020;                     /* close block */    
               if(liflags & FILEWRT){
                  for(j=0; j<=count; j++)     /* use cached indices */
                       saveliteval(bind,(short)((lvp0+j)->val[0]),lbk->bldep,lbk->lvarr+j);    
               }
	   }
        }
   }	                     
   else{                     /* normal procedure */

    if(count >= lbk->maxlvs){
        if(addlvstore(bind, lbk->bldep) == 0){                                                                                                                                        
	         return(0);
        }		                                                                            
    }                           /* insert */                                          	   
    lvp=lbk->lvarr+count;                                
    for(j=0; j<3; j++)
           lvp->dir[j] = (float)nrm[j];
    copycolor(lvp->val,lcol);
    lbk->indblock[sind] = count;    
#ifdef INFO
    setlvs++;
#endif    
    if(liflags & FILEWRT){         
     saveliteval(bind,sind,lbk->bldep,lvp);
    }
    }
 break;
 
 case 0:                       /* alloc new block and store first val*/
    if(initlvstore(bind, sdep) == 0)
                               return(0);
    sind = getsindex(delpos,dcsubd);   /* get subblockindex */
    lbk=liblkptr[bind];
    lvp=lbk->lvarr;            /*insert */
    for(j=0; j<3; j++)
       lvp->dir[j] = (float)nrm[j];
    copycolor(lvp->val,lcol);                                                                             
    lbk->indblock[sind] = 0;
#ifdef INFO
    setlvs++;
#endif    
    lvp+= dctval;           /* cache position/index/brightness in val-array */  
    for(j=0; j<3; j++)        
         lvp->dir[j] = delpos[j]; 
    (short)(lvp->val[0]) = sind;                 
    lvp->val[1] = 0.265*lcol[0] + 0.670*lcol[1] + 0.065*lcol[2];
    (lvp+1)->val[2] = lvp->val[1] + FTINY;
    lvp->val[2] = 0;
 break;
 }/* end switch */ 
 return(1);
}

int
setdrclite(COLOR lcol, RAY *r, FVECT nrm)  
{                              
 int bind;
 short sind;
 register LITEBLOCK *lbk;

 bind = getbindex(r);        /* determine position in octree */ 
 
 if(liblkptr[bind] != NULL){            

	lbk=liblkptr[bind]; 
	sind = getsindex(delpos, (lbk->bldep & 017));
	
        if(lbk->indblock[sind] != -1)
                              return(0);
	else
	  setlite(lcol,nrm,bind,sind,1);
	  return(1);		      
 }
 else{
        sind = getsindex(delpos,dcsubd); 
 	setlite(lcol,nrm,bind,sind,0);		      
 }          
 return(1);  
}

int                          /* alloc one liteblock with space for its arrays*/
initlvstore(int bind, char dp)
{
   register int i;
   int rda;
   char dep;
   
   dep = dp & 017;
   
   rda = rdarr[dep];                
                                                                       
   liblkptr[bind] = (LITEBLOCK*)malloc(sizeof(LITEBLOCK));
     if(liblkptr[bind] == NULL){
            error(WARNING, "DC: out of memory");
            return(0);
     }
   
   liblkptr[bind]->bldep = dp;
   
   liblkptr[bind]->indblock = (short*)malloc(aloc[dep]*sizeof(short));
     if(liblkptr[bind]->indblock == NULL){
            error(WARNING, "DC: out of memory");
            return(0);
     }
   for(i=0; i<aloc[dep]; i++)         /* init to -1 to indicate emptyness */
                 *(liblkptr[bind]->indblock+i) = -1;
   
   
   liblkptr[bind]->lvarr =(LITEVAL*)malloc((rda)*LITEVALSIZ);
     if(liblkptr[bind]->lvarr == NULL){
            error(WARNING, "DC: out of memory");
            return(0);
     }
      
   liblkptr[bind]->maxlvs = rda;  /* initialize */        
   liblkptr[bind]->nlvs = 0;         

#ifdef INFO   
   blockalc++;       
   totmem+= sizeof(LITEBLOCK) + aloc[dcsubd]*sizeof(short) + rda*LITEVALSIZ;
#endif   
   return(1);                          
}    
 
int
addlvstore(int bind, char bdep)
{
register int newmax;
#ifdef INFO
register int del;
#endif
register int i;
 int rda;
 LITEVAL* newlvarr;
 char dep;
 
 dep = bdep & 017;
                     /* set new reduced storage number */
 rda = (rdarr[dep]>>1);

 newmax = liblkptr[bind]->maxlvs + rda;

 if (newmax > aloc[dep])
           newmax = aloc[dep];
#ifdef INFO  
 del = newmax-liblkptr[bind]->maxlvs;
#endif

 liblkptr[bind]->maxlvs = newmax;
                                 /* selfmade reallocation - the radiance-malloc/bmalloc-versions*/
				 /* seem to be incompatible with the library-realloc */
 newlvarr = (LITEVAL*)malloc(newmax*LITEVALSIZ);
 if(newlvarr == NULL){
            error(WARNING, "DC: out of memory");
            return(0);
 }
 for(i=0; i<liblkptr[bind]->maxlvs; i++)
    *(newlvarr+i) = *(liblkptr[bind]->lvarr+i);
  
 free(liblkptr[bind]->lvarr);       /* exchange */
 liblkptr[bind]->lvarr = newlvarr;  
#ifdef INFO  
 totmem+= del*LITEVALSIZ;  
 blocksrealc++;
#endif
return(1);
}

int
reablk(int bind, char dp, int cnt)
{
  register int i;
#ifdef INFO
 register int del;
#endif  
  short* newindblock;
  LITEVAL* newlvarr;
  int nwrda;
  char dep;
  
  dep = dp & 017;
                       /* change of blockdepth, so new indexlist needed */
  liblkptr[bind]->bldep = dp;
  
  newindblock = (short*)malloc(aloc[dep]*sizeof(short));
  if(newindblock == NULL){
            error(WARNING, "DC: out of memory");
            return(0);
  }     
  for(i=0; i<aloc[dep]; i++)       /* init to -1 to indicate emptyness */
        *(newindblock+i) = -1;
      
 free(liblkptr[bind]->indblock);      /* exchange */
 liblkptr[bind]->indblock = newindblock;
 
                                    /* set new reduced storage number */
  nwrda = rdarr[dep];

#ifdef INFO      
  del = aloc[dep] - aloc[dcsubd];
  totmem+= del*sizeof(short);
  del = nwrda-liblkptr[bind]->maxlvs;
#endif      

  liblkptr[bind]->maxlvs = nwrda;
                        /* selfmade reallocation - the radiance-malloc/bmalloc-versions*/
		        /* seem to be incompatible with the library-realloc */
  newlvarr = (LITEVAL*)malloc(nwrda*LITEVALSIZ);
  if(newlvarr == NULL){
            error(WARNING, "DC: out of memory");
            return(0);
  }
                                /* shift existing vals */
 for(i=0; i<=(dctval+cnt); i++)
    *(newlvarr+i) = *(liblkptr[bind]->lvarr+i);
  
 free(liblkptr[bind]->lvarr);
 liblkptr[bind]->lvarr = newlvarr;  

#ifdef INFO  
 totmem+= del*LITEVALSIZ;  
#endif
                    
return(1);
}                    
      
    
                                                       
                  